; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         mh-alias.el
; RCS:          $Header: mh-alias.el,v 1.4 93/01/26 14:20:06 darrylo Exp $
; Description:  MH alias handling routines
; Author:       Darryl Okahata
; Created:      Tue Sep 24 18:01:15 1991
; Modified:     Wed Jul 28 11:08:46 1993
; Language:     Emacs-Lisp
; Package:      N/A
; Status:       Experimental
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; This file provides optional MH alias expansion for mh-e.el.
;;
;; When you are prompted for one or more addresses, or when the cursor
;; is within a To:, Bcc:, or Cc: mail header, pressing TAB will cause
;; the alias just before the cursor to be expanded, if it exists and is
;; unique.  If the alias is not unique, all matching possibilities are
;; displayed in a pop-up buffer.  If there is no alias before the
;; cursor, all aliases are displayed.
;;
;; In order to have alias completion, you must have a set of valid MH
;; aliases, as the "ali" command is used to get a list of MH aliases.
;; By setting the variable `mh-read-aliases-from-passwd' to non-nil, all
;; usernames in the file given by `mh-etc-passwd-file' ("/etc/passwd" by
;; default) will be added to the list of aliases.  Note that this
;; requires a readable password file in order to work.
;;
;; Configuration variables:
;;
;;	mh-read-aliases-from-passwd
;;		If non-nil, merge usernames from the password file given
;;		by `mh-etc-passwd-file' into the list of MH aliases.
;;	mh-use-system-aliases
;;		If non-nil, read aliases from sendmail-style alias
;;		files.  The list of sendmail-style alias files is given
;;		by `mh-system-aliases-file-list'.  Note that merging
;;		system aliases can take a while.  On my moderately fast
;;		workstation, it takes just over a minute to rebuild the
;;		list of aliases (however, my site has about 3000
;;		aliases).
;;	mh-alias-fully-expand-aliases
;;		If non-nil, allows mh-expand-alias to replace a
;;		fully-specified and unambiguous alias by its expansion.
;;		If nil, aliases are expanded only up to the point of
;;		being fully-specified and unambiguous; the alias name is
;;		not replaced with its expansion.
;;	mh-etc-passwd-file
;;		The /etc/passwd file to use to read usernames.
;;	mh-system-aliases-file-list
;;		List of sendmail-style alias files to scan, if system
;;		aliases are used.  Some sites have multiple sendmail
;;		alias files, which is why this variable is a list.
;;	mh-ask-to-address
;;	mh-ask-cc-address
;;	mh-ask-subject
;;		If nil, suppress the asking of the corresponding header info,
;;		when composing, forwarding, or redistributing mail.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'mh-e)			;; mh-e.el must be loaded before this one

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; User-settable variables (don't change them here -- change them in
;; your .emacs file):
;;

(defvar mh-read-aliases-from-passwd nil
  "*If non-nil, merge usernames from the password file given by
`mh-etc-passwd-file' into the list of MH aliases.")


(defvar mh-use-system-aliases nil
  "*If non-nil, read aliases from sendmail-style alias files.
The list of sendmail-style alias files is given by
`mh-system-aliases-file-list'.  Note that merging system aliases can
take a while.  On my moderately fast workstation, it takes just over a
minute to rebuild the list of aliases (however, my site has about 3000
aliases).")


(defvar mh-alias-fully-expand-aliases nil
  "*If non-nil, allows mh-expand-alias to replace a fully-specified and
unambiguous alias by its expansion.  If nil, aliases are expanded only
up to the point of being fully-specified and unambiguous; the alias name
is not replaced with its expansion.")


(defvar mh-etc-passwd-file "/etc/passwd"
  "The /etc/passwd file to use to read usernames.")


(defvar mh-system-aliases-file-list '("/usr/lib/aliases")
  "List of sendmail-style alias files to scan, if system aliases are
used.  Some sites have multiple sendmail alias files, which is why this
variable is a list.")


(defvar mh-ask-to-address t
  "*If non-nil, ask the user for a To: address when composing or forwarding
a message.")


(defvar mh-ask-cc-address t
  "*If non-nil, ask the user for a Cc: address when composing or forwarding
a message.")


(defvar mh-ask-subject t
  "*If non-nil, ask the user for a Subject: when composing or forwarding
a message.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; These variables are not intended for modification by the user.
;;

(defvar mh-aliases-list nil
  "List of known mh aliases.
Generated at run time from various sources (e.g., the MH `ali' command).")


(defvar mh-read-address-keymap
  (let ( (map (copy-keymap minibuffer-local-map)) )
    (define-key map "\t" 'mh-expand-alias)
    map
    )
  "Keymap to use when reading email addresses/aliases.  The default is to
use the existing minibuffer-local-map as a template.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun mh-read-etc-passwd (file)
  "Read usernames from the password file FILE.
The usernames are returned in a form suitable for completing-read.
If the file cannot be read, nil is returned; an error is not raised."
  (let (result (temp-buf " *mh-temp*") uname)
    (save-excursion
      (set-buffer (mh-get-buffer temp-buf))
      (erase-buffer)
      (condition-case err
	  (insert-file-contents (expand-file-name
				 (substitute-in-file-name file)))
	(error nil))	;; ignore errors
      (goto-char (point-min))
      (while (looking-at "^\\([^:]+\\):")
	(setq uname (buffer-substring (match-beginning 1)
				      (match-end 1)))
	(setq result (cons (cons uname uname) result))
	(forward-line 1)
	)
      )
    result
    ))


(defun mh-string-in-list (str list)
  "Check to see if string (STR) is in LIST."
  (let ()
    (catch 'exit
      (while list
	(if (string= (car str) (car (car list)))
	    (throw 'exit t))
	(setq list (cdr list))
	))
    ))


(defun mh-make-alias-list (&optional in-minibuffer list)
  "Return a list of the user's MH aliases.
Result is in a form suitable for completing read.
If `in-minibuffer is non-nil, the current buffer is the minibuffer, and
messages will be displayed using an alternative scheme, one which does not
interfere with the display of the minibuffer.."
  (interactive)
  (let ( (case-fold-search t)
	 original-contents origin alias expansion
	 )
    (unwind-protect
	(progn
	  (if in-minibuffer
	      (progn
		(setq origin (point)
		      original-contents (buffer-substring (point-min)
							  (point-max)))
		(erase-buffer)
		(insert " [ Collecting aliases ... ]")
		(sit-for 0)
		)
	    (message "Collecting aliases ..."))
	  (save-window-excursion
	    (mh-exec-cmd-quiet " *mh-temp*"
			       "ali" "-nolist" "-nonormalize" "-nouser")
	    (goto-char (point-min))
	    
	    (while (re-search-forward "^\\([-\\+a-z0-9_]+\\):\\s-+\\(.*\\)$" nil t)
	      (setq alias (buffer-substring (match-beginning 1) (match-end 1))
		    expansion (buffer-substring (match-beginning 2)
						(match-end 2)))
	      (if (not (assoc alias list))
		(progn
		  (forward-line 1)
		  (while (looking-at "\\s-")
		    (setq expansion (concat expansion
					    (buffer-substring (1- (point))
							      (progn
								(forward-line 1)
								(1- (point)))))))
		  (setq list (cons (cons alias expansion) list))))
	      )
	    list))
      (progn
	(if in-minibuffer
	    (progn
	      (erase-buffer)
	      (insert original-contents)
	      (goto-char origin)
	      )
	  (message "Collecting aliases ... done"))
	)
      )))


(defun mh-make-passwd-alias-list (&optional in-minibuffer list)
  "Return a list of aliases, obtained from the passwd file.
The passwd file to examine is specified by `mh-etc-passwd-file'
variable."
  (interactive)
  (let (plist minibuf origin original-contents)
    (unwind-protect
	(progn
	  (if in-minibuffer
	      (progn
		(setq minibuf (current-buffer)
		      origin (point)
		      original-contents (buffer-substring (point-min)
							  (point-max)))
		(erase-buffer)
		(insert " [ Collecting passwd aliases ... ]")
		(sit-for 0)
		)
	    (message "Collecting passwd aliases ... "))
	  (setq plist (mh-read-etc-passwd mh-etc-passwd-file))
	  (while plist
	    (if (not (assoc (car (car plist)) list))
		(setq list (cons (car plist) list)))
	    (setq plist (cdr plist))
	    )
	  )
      (progn
	(if (and in-minibuffer minibuf original-contents)
	    (progn
	      (set-buffer minibuf)
	      (erase-buffer)
	      (insert original-contents)
	      (goto-char origin)
	      )	
	  (message "Collecting passwd aliases ... done"))
	))
    list
    ))


(defun mh-make-system-alias-list (&optional in-minibuffer list)
  "Return an alist of the system-wide aliases, similar to the above function."
  (interactive)
  (let ( (file-list mh-system-aliases-file-list)
	 alias aliases origin expansion minibuf original-contents
	 )
    (unwind-protect
	(progn
	  (save-window-excursion
	    (if in-minibuffer
		(progn
		  (setq minibuf (current-buffer)
			origin (point)
			original-contents (buffer-substring (point-min)
							    (point-max)))
		  (erase-buffer)
		  (insert " [ Collecting system aliases (this can take a while) ... ]")
		  (sit-for 0)
		  )
	      (message "Collecting system aliases (this can take a while) ... ")
	      )
	    (buffer-disable-undo (set-buffer
				  (setq aliases
					(get-buffer-create "*system-aliases"))))
	    (erase-buffer)
	    (goto-char (point-min))
	    (while file-list
	      (insert-file-contents (expand-file-name
				     (substitute-in-file-name
				      (car file-list))))
	      (goto-char (point-max))
	      (setq file-list (cdr file-list))
	      )
	    (goto-char (point-min))
	    (while (re-search-forward
		    "^\\([-_a-z0-9]+\\)\\s-*:\\s-*\\(.*\\)$" nil t)
	      (setq alias (buffer-substring (match-beginning 1) (match-end 1))
		    expansion (buffer-substring (match-beginning 2)
						(match-end 2)))
	      (if (assoc alias list)
		  nil
		(forward-line 1)
		(while (looking-at "[ \t]")
		  (setq expansion (concat expansion
					  (buffer-substring (1- (point))
							    (progn
							      (forward-line 1)
							      (1- (point)))))))
		(setq list (cons (cons alias expansion) list))))
	    (kill-buffer aliases)
	    list))
      (progn
	(if (and in-minibuffer minibuf original-contents)
	    (progn
	      (set-buffer minibuf)
	      (erase-buffer)
	      (insert original-contents)
	      (goto-char origin)
	      ))
	(message "Collecting system aliases (this can take a while) ... done")
	)
      )))


(defun mh-rebuild-aliases (in-minibuffer)
  "Rebuild the list of aliases to be used for completion."
  (interactive '(nil))
  (let (aliases plist)
    ;;
    ;; System aliases are created first, so that these aliases will end
    ;; up a the END of the list.  We want system aliases searched after
    ;; MH aliases.
    ;;
    (if mh-use-system-aliases
	(setq aliases (mh-make-system-alias-list in-minibuffer aliases)))
    ;;
    ;; By checking for passwd aliases here, we insure that passwd
    ;; aliases are added before MH aliases.  We want any MH aliases to
    ;; take precedence over any in the passwd file.
    ;;
    (if mh-read-aliases-from-passwd
	(setq aliases (mh-make-passwd-alias-list in-minibuffer aliases)))
    ;;
    ;; Now, merge in MH aliases.
    ;;
    (setq aliases (mh-make-alias-list in-minibuffer aliases))
    (setq mh-aliases-list aliases)
    ))


(defun mh-string< (a b)
  (let ()
    (setq a (downcase a)
	  b (downcase b))
    (string< a b)
    ))


(defun mh-pop-up-aliases (aliases)
  "Display to the user a list of aliases."
  (let (display-aliases)
    (save-excursion
      (if (not (stringp (car aliases)))
	  (while aliases
	    (setq display-aliases (cons (car (car aliases)) display-aliases))
	    (setq aliases (cdr aliases))
	    )
	(setq display-aliases (copy-sequence aliases)))
      (setq display-aliases (sort display-aliases 'mh-string<))
      (with-output-to-temp-buffer
	  " *Completions*"
	(display-completion-list display-aliases))
      )))


(defun mh-handle-letter-mode-key (key)
  (let ()
    (cond
     ( (or (eq key 'tab) (and (numberp key) (= key ?\t))) (tab-to-tab-stop) )
     ( t (insert key) )
     )))


(defun mh-expand-alias (rebuild-aliases)
  "Expand the alias before point.
In order for the alias to be expanded, point must be either in the minibuffer
or in the To: or Cc: mail headers.  Otherwise, the keypress used to invoke
this function is instead inserted into the buffer.

If the alias is not unique, all matching possibilities are displayed in
a pop-up buffer.  If there is no alias before point, all aliases are
displayed.

If the alias is unique, it is replaced with its expanded value if the
`mh-alias-fully-expand-aliases' variable is non-nil."
  (interactive "P")
  (let ( (origin (point))
	 (case-fold-search t)
	 abbrev aliases in-minibuffer)
    (if (string-match "^ [*]Minibuf-[0-9]" (buffer-name))
	(progn
	  (setq in-minibuffer t)
	  )
      (progn
	(beginning-of-line)
	(while (and (< (point-min) (point))
		    (looking-at "[ \t]")
		    )
	  (forward-line -1))
	))
    (cond
     ( (or in-minibuffer
	   (looking-at "To:\\|Cc:\\|Bcc:")
	   )
       (if (or (not mh-aliases-list) rebuild-aliases)
	   (mh-rebuild-aliases in-minibuffer))
       (goto-char origin)
       (if (and (re-search-backward "\\(^\\)\\|\\([,: \t]\\)" nil t)
		(not (bolp)))
	   (forward-char 1))
       (if (or (and (looking-at "[ \t,:]*\\([-_a-z0-9]+\\)")
		    (not (bolp)))
	       in-minibuffer
	       (looking-at "[ \t,:\n]")
	       )
	   (let ( (start (match-beginning 1))
		  (stop (match-end 1))
		  (alias-list mh-aliases-list)
		  )
	     (if (and start stop (/= start stop) alias-list)
		 (progn
		   (setq abbrev (buffer-substring start stop))
		   (setq aliases (try-completion abbrev alias-list))
		   (cond ((eq aliases t)	; found a match: replace it!
			  (if (not mh-alias-fully-expand-aliases)
			      (goto-char stop)
			    (let ( (replacement (cdr (assoc abbrev
							    alias-list))) )
			      (if (string= abbrev replacement)
				  (progn
				    (goto-char stop)
				    (unwind-protect
					(progn
					  (insert " [ alias is unique ]")
					  (ding)
					  (sit-for 3)
					  )
				      (delete-region origin (point)))
				    )
				(progn
				  (delete-region start stop)
				  (goto-char start)
				  (if (not (or (= (preceding-char) ? )
					       (= (preceding-char) ?\t)
					       (bolp)))
				      (insert " "))
				  (insert replacement)))
			      )))
			 ((eq aliases nil)	; no such match: leave alone
			  (goto-char stop)
			  (unwind-protect
			      (progn
				(insert " [No Match]")
				(ding)
				(sit-for 3)
				)
			    (delete-region origin (point)))
			  )
			 ((string= abbrev aliases)
			  (goto-char stop)
			  (mh-pop-up-aliases
			   (all-completions abbrev alias-list)))
			 ((stringp aliases)
			  (delete-region start stop)
			  (goto-char start)
			  (if (not (or (= (preceding-char) ? )
				       (= (preceding-char) ?\t)
				       (bolp)))
			      (insert " "))
			  (insert aliases))
			 ))
	       (progn
		 (goto-char origin)
		 (mh-pop-up-aliases alias-list)
		 )))
	 (progn
	   (goto-char origin)
	   (mh-handle-letter-mode-key last-input-char)
	   )))
     ( (looking-at "Fcc:")
       (goto-char origin)
       (error "Please use %s for Fcc: completion"
	      (key-description (where-is-internal 'mh-to-fcc
						  mh-letter-mode-map t t)))
       )
     ( t
       (goto-char origin)
       (mh-handle-letter-mode-key last-input-char)
       )
     )))


(defmacro mh-read-address (prompt)
  "Stupid macro to read a list of MH addresses."
  (` (read-from-minibuffer (, prompt) nil mh-read-address-keymap) ))


(defun mh-read-to-address (&optional prompt)
  "Ask the user for a To: address only if mh-ask-to-address is non-nil."
  (if (not prompt)
      (setq prompt "To: "))
  (if mh-ask-to-address
      (mh-read-address prompt)
    ""))


(defun mh-read-cc-address (&optional prompt)
  "Ask the user for a Cc: address, only if mh-ask-to-address is non-nil."
  (if (not prompt)
      (setq prompt "Cc: "))
  (if mh-ask-cc-address
      (mh-read-address prompt)
    ""))


(defun mh-read-subject (&optional prompt)
  "Ask the user for a Subject: only, if mh-ask-to-address is non-nil."
  (if (not prompt)
      (setq prompt "Subject: "))
  (if mh-ask-subject
      (read-from-minibuffer prompt)
    ""))


(defun mh-read-to-cc-subject ()
  "Read addresses for To: and Cc:, and read a subject line.
The result is returned in a form suitable for passing to the `interactive'
function.
This should really be a macro"
  (list (mh-read-to-address) (mh-read-cc-address) (mh-read-subject)))
