; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         mh-shar.el
; RCS:          $Header: mh-shar.el,v 1.2 93/01/20 11:13:36 darrylo Exp $
; Description:  shar/unshar/uuencode/uudecode support for mh-e.el.
;		Also contains MD4 message message integrity checking code.
;		(This should really be placed somewhere else.)
; Author:       Darryl Okahata
; Created:      Tue Sep 24 18:38:14 1991
; Modified:     Fri Nov  8 12:23:34 1991 (Darryl Okahata) darrylo@hpsrdmo
; Language:     Emacs-Lisp
; Package:      N/A
; Status:       Experimental
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'mh-e)			;; mh-e.el must be loaded before this one

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; This file provides shar/unshar, uuencode/uudecode, and MD4 message
;; verification support.
;;
;; See the variables for configuration options.
;;
;; The MH scan buffer has these additional default keybindings:
;;
;;	M-n	mh-unshar
;;		Unshar the current message.  You will be prompted for
;;		the name of a directory in which to unshar (the
;;		directory will be created if necessary).
;;	M-d	mh-uudecode
;;		Uudecode the current message.  You will be prompted for
;;		the name of a directory in which to unshar (the
;;		directory will be created if necessary).
;;	M-v	mh-check-md4-signature
;;		*IF* the current message contains an MD4 signature in
;;		the header, this function will verify the current
;;		message against the given MD4 signature.  Note that you
;;		need to have the comp.sources.misc MD4 verification
;;		programs in order for this to work.  An MD4 signature
;;		will automatically be added to outgoing messages if
;;		`mh-add-md4-signature' is non-nil.
;;
;;		***** IMPORTANT: you need MH 6.7.1 or later in order for
;;		this to work, as there is a bug in earlier versions that
;;		slightly corrupts mail messages (it adds an extra
;;		newline to messages, which changes the MD4 signature).
;;		Also, if you use Chip Salzenberg's deliver program, you
;;		need version 2.1 or later, as earlier versions have a
;;		bug that corrupts messages with MD4 signatures in the
;;		header.
;;
;; The MH letter composition buffer has these additional default
;; keybindings:
;;
;;	M-s	mh-shar-and-insert
;;		Prompt for a file/directory, and insert, at point, the
;;		file/directory as a shell archive.  You are also
;;		prompted for additional shar command line options; this
;;		is primarily intended for those shar programs that
;;		support optional compression of files.  However, if you
;;		have more than one file to archive, you can specify all
;;		but the last one when prompted for the additional
;;		options; give the last name when asked for the
;;		file/directory to shar.
;;	M-u	mh-uuencode-and-insert
;;		Prompt for a file, and insert that file as a uuencoded
;;		object.  If a prefix argument is given, the uuencoded
;;		object will contain an absolute pathname.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; unshar and uudecode gunk:
;;


(defvar mh-last-extract-directory nil
  "The last directory used to unshar or uudecode a message.")


(defvar mh-uudecode-program "uudecode"
  "The name of the program used to uudecode a message.	The file to uudecode
is piped to this program, and no arguments are passed to it.")

(defvar mh-uuencode-program "uuencode"
  "The name of the program used to uuencode a message.	The file to uuencode
is piped to this program, and the same filename is given as the argument.")


(defvar mh-shar-program "shar"
  "Program to use to create a shar archive.")

(defvar mh-shar-standard-program-args
  (cond ((eq system-type 'hpux)
	 '("-C" "-c" "-s"))
	(t nil))
  "Standard set of arguments to pass to the shar program.")


(if (not (fboundp 'make-directory))
    (progn
      (defun make-directory (fn)
	"Make a directory.
This is a subr in Emacs 19.
This is here for backwards V19 compatibility."
	(interactive (list (read-file-name "Make directory: ")))
	(setq fn (expand-file-name fn))
	(if (file-exists-p fn)
	    (error "Cannot make directory %s: file already exists" fn)
	  (call-process "mkdir" nil nil nil fn))
	(or (file-directory-p fn)
	    (error "Could not make directory %s" fn)))
      ))


(defun mh-multiple-regexp-search (regexps)
  "Simultaneously search forward for multiple regexps.
If a match is found, point is left at the end of the match closest to
the starting position, and `t' is returned.  If a match is not found,
`nil' is returned, and point is not moved..  REGEXPS is a list of
regular expressions to search for."
  (let ( (l regexps)
	 (case-fold-search t)
	 (start (point))
	 item
	 match-loc
	 )
    (while l
      (setq item (car l))
      (goto-char start)
      (if (re-search-forward item match-loc t)
	  (setq match-loc (point)))
      (setq l (cdr l))
      )
    (if match-loc
	(progn
	  (goto-char match-loc)
	  t)
      (progn
	(goto-char start)
	nil))
    )
  )


(defun mh-read-and-create-directory ()
  "Prompt the user for the name of a directory.
If the directory does not already exist, ask the user if he wants to create it.
If he does, the directory is created; if he doesn't, an error is raised.
Note that the last directory name is kept in `mh-last-extract-directory'.
This variable is updated by the function that calls this one, so that the
variable is updated only if the operation is successful.

This function is intended to be used to create an argument list for the
`interactive' function, e.g.:

	(defun foofun (directory)
	  (interactive (mh-read-and-create-directory))
	  ...
	)
"
  (let (dir)
    (if (or (not mh-last-extract-directory)
	    (not (stringp mh-last-extract-directory))
	    (not (file-directory-p mh-last-extract-directory))
	    )
	(setq mh-last-extract-directory default-directory))
    (setq dir (read-file-name "Unshar/uudecode in directory: "
			      mh-last-extract-directory
			      mh-last-extract-directory))
    ;; set `dir' to absolute, expanded path name
    (setq dir (directory-file-name (expand-file-name
				    (substitute-in-file-name dir))))
    (if (and (file-exists-p dir)
	     (not (file-directory-p dir)))
	(error (format "`%s' is a regular file, not a directory" dir)))
    (if (not (file-directory-p dir))
	(progn
	  (if (y-or-n-p (format "Make directory %s? " dir))
	      (make-directory dir)
	    (error "Directory %s not made" dir)
	    )
	  )
      )
    (cons dir nil)
    ))


(defun mh-do-unshar (dir)
  "Do the actual unshar for `mh-unshar'.
This function is not really mh-e specific; you should be able to execute
this function in any buffer that contains a shar file."
  (interactive (mh-read-and-create-directory))
  (let ( (errbuf (get-buffer-create " *Unshar Output*"))
	 )
    (message "Looking for start of shar package ...")
    (goto-char (point-min))
    ;;
    ;; Here, we MUST search for the FIRST occurrence of the regexps:
    ;;
    ;;		^#[ \t]+This[ \t]+is[ \t]+a[ \t]+shell[ \t]+archive
    ;;		^#![ \t]*/bin/sh
    ;;
    ;; We must do this because some crazy people may create shar
    ;; archives of shar archives.  In this case, the shar archive
    ;; may not start with "#! /bin/sh", but may contain items that
    ;; do.  Yuk.
    ;;
    ;; Should we add a "cut here" regexp to the below list?
    ;;
    (if (or (mh-multiple-regexp-search
	     '("^#[ \t]+This[ \t]+is[ \t]+a[ \t]+shell[ \t]+archive"
	       "^#![ \t]*/bin/sh"))
	    (and (re-search-forward "^[^a-z0-9\"]*cut[ \t]+here" nil t)
		 (forward-line 1))
	    (re-search-forward "^#" nil t)
	    (re-search-forward "^: " nil t)
	    )
	(let ( (start (progn (beginning-of-line) (point)))
	       (default-directory (file-name-as-directory dir))
	       )
	  (save-excursion
	    (set-buffer errbuf)
	    (buffer-flush-undo errbuf)
	    (erase-buffer)
	    (insert "Unsharing in directory: \"" dir "\"\n\n")
	    )
	  (message "Please wait ...")
	  (set-window-start (display-buffer errbuf) 0) ;so can watch progress
	  (call-process-region start (point-max)
			       "/bin/sh" nil errbuf t)
	  (message "Unshar done")
	  (setq mh-last-extract-directory dir)
	  )
      (error "Can't find start of shar file.")
      )
    )
  )


(defun mh-external-function (dir function)
  ""
  (let ()
    (mh-show nil)		;;; force showing of current message
    (save-excursion
      (set-buffer (mh-get-buffer mh-show-buffer))
      (buffer-flush-undo (current-buffer))
      (funcall function dir)
      )
    ))


(defun mh-unshar (dir)
  "Unshar the current message in the directory given by DIR."
  (interactive (mh-read-and-create-directory))
  (let ()
    (mh-external-function dir 'mh-do-unshar)
    ))


(defun mh-do-uudecode (dir)
  "Do the actual uudecode for `mh-uudecode'.
This function is not really mh-e specific; you should be able to execute
this function in any buffer that contains a uuencoded file."
  (interactive (mh-read-and-create-directory))
  (let ( (default-directory (file-name-as-directory dir)) )
    (message "Uudecoding in directory `%s' ..." dir)
    (call-process-region (point-min) (point-max) mh-uudecode-program nil)
    (message "Uudecoding in directory `%s' ... done" dir)
    (setq mh-last-extract-directory dir)
    ))


(defun mh-uudecode (dir)
  "Uudecode the current message in the directory given by DIR."
  (interactive (mh-read-and-create-directory))
  (let ()
    (mh-external-function dir 'mh-do-uudecode)
    ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Shar and uuencode gunk:
;;

(defun mh-shar-and-insert (file other-args)
  "Shar the file FILE, passing OTHER-ARGS (if any) to the shar program.
The shar contents in then inserted into the current buffer."
  (interactive "DFile or directory to shar: \nsOther args: ")
  (let ((start (point)))
    (setq file (expand-file-name (substitute-in-file-name file)))
    (apply 'call-process mh-shar-program nil t nil
	   (if (and other-args (not (string= other-args "")))
	       (append mh-shar-standard-program-args (mh-split-args other-args)
		       (cons file nil))
	     (append mh-shar-standard-program-args (cons file nil))
	     ))
    (goto-char start)	;; go back to starting position
    ))


(defun mh-uuencode-and-insert (file &optional absolute)
  "Uuencode and insert the file FILE into the current buffer.
Normally, the uuencoded filename does not contain a pathname.
If ABSOLUTE is non-nil, or if a prefix argument is given, the uuencoded
information will contain an absolute pathname."
  (interactive "FFile to uuencode and insert: \nP")
  (let ((start (point)) uuencode-filename)
    (setq file (expand-file-name (substitute-in-file-name file)))
    (if absolute
	(setq uuencode-filename file)
      (setq uuencode-filename (file-name-nondirectory file)))
    (call-process mh-uuencode-program file t nil uuencode-filename)
    (goto-char start)	;; go back to starting position
    ))


;; Stolen from rcs.el:
(defun mh-split-args (args)
  "Splits the string ARGS into a list of strings, one argument per."
  (let (result)
    (if (null args)
	nil
      (while (> (length args) 0)
	(let ((end-arg (string-match "[ \t]+" args)))
	  (if end-arg
	      (progn
		(setq result (cons (substring args 0 end-arg) result))
		(setq args (substring args (match-end 0))))
	    (setq result (cons args result))
	    (setq args ""))))
      (nreverse result))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MD4 message integrity gunk:
;;

(defvar mh-add-md4-signature nil
  "*Controls the addition of MD4 signatures to the outgoing message header.
If non-nil, a valid \"X-Md4-Signature:\" header line is added to outgoing
mail messages.")

;;; Variables used when mh-add-md4-signature is non-nil:

(defvar mh-md4-program "md4"
  "The MD4 program.
Text piped to stdin should produce an MD4 signature on stdout.")

(defvar mh-checkmd4-program "checkmd4"
  "The MD4 program that checks a message for a valid MD4 signature.
This program should write a descriptive string to stdout saying whether or
not the message, piped to stdin, matches the MD4 signature in the message
header.")

(defvar mh-checkmd4-valid-string "Valid."
  "The string printed by the mh-checkmd4-program if the signature is valid.
This string is considered to be anchored to the beginning of the first line
printed by the mh-checkmd4-program.")


(defun mh-add-md4-header ()
  "Insert an \"X-Md4-Signature:\" line into the header."
  (let (
	(outbuf " *md4*")
	checksum
	)
    (save-excursion
      (set-buffer (get-buffer-create outbuf))
      (buffer-flush-undo (current-buffer))
      (kill-region (point-max) (point-min)))
    (save-excursion
      (goto-char (point-min))
      (if (re-search-forward "^--------" nil t)
	  (progn
	    (forward-line 1)
	    (call-process-region (point) (point-max) mh-md4-program nil outbuf)
	    (save-excursion
	      (set-buffer outbuf)
	      (goto-char (point-min))
	      (end-of-line)
	      (setq checksum (buffer-substring (point-min) (point)))
	      )
	    (mh-insert-header-line (format "X-Md4-Signature: %s\n" checksum))
	    )
	)
      )
    ))


(defun mh-check-md4-signature ()
  "If an MD4 signature is present in the header, check it and report the results."
  (interactive)
  (let (start end checksum (outbuf " *md4*"))
    (mh-show nil)		;;; force showing of current message
    (save-excursion
      (mh-set-buffer mh-show-buffer)
      (goto-char (point-min))
      (if (re-search-forward "\\(^$\\)\\|\\(^--------\\)" nil t)
	  (progn
	    (setq end (point))
	    (goto-char (point-min))
	    (if (re-search-forward "^X-Md4-Signature:[ \t]*" end t)
		(progn
		  (save-excursion
		    (set-buffer (get-buffer-create outbuf))
		    (kill-region (point-max) (point-min)))
		  (call-process-region (point-min) (point-max)
				       mh-checkmd4-program nil outbuf)
		  (set-buffer outbuf)
		  (goto-char (point-min))
		  (if (looking-at mh-checkmd4-valid-string)
		      (message "Message is uncorrupted (MD4 signature is OK).")
		    (error "MD4 signature does not match -- possible message corruption!"))
		  )
	      (progn
		(error "This message has no MD4 signature!")
		)
	      )
	    )
	(progn
	  (error "Unable to find end of message header!")
	  )
	)
      )
    )
  )
