From: dodd@roebling.poly.edu (Lawrence R. Dodd)
Subject: tex-narrow.el: Additional narrowed region handling for AUCTeX.
Date: 17 Apr 1994 15:04:37 -0400
Sender: daemon@cis.ohio-state.edu


`C-c n n' will narrow to current chapter, section, subsection, or
document environment in a LaTeX document.  This makes it easier to
concentrate on just one part of a large document.

`C-c C-v' will run TeX on the current narrowed region.  This is like
a specialized version `C-c C-r' (TeX-command-region) but, in my
opinion, easier to use.

Requires AUCTeX.  See /ftp.iesd.auc.dk:/pub/emacs-lisp/

---
;;; tex-narrow.el -- Additional narrowed LaTeX region handling within AUCTeX.

;; Author: Lawrence R. Dodd <dodd@roebling.poly.edu>
;; Maintainer: Lawrence R. Dodd <dodd@roebling.poly.edu>
;; Created: Mon Apr 11 12:42:46 1994
;; Version: 1.10
;; Date: 1994/04/17 18:57:05
;; Keywords: AUCTeX extensions

;;; Copyright (C) 1994  Lawrence R. Dodd
;;;  
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:
;;; 
;;; Code for working with narrowed regions within AUCTeX. 
;;; Binds `C-c C-v' and `C-c n n' in LaTeX.

(defun TeX-command-narrowed-region ()
  "Run TeX on narrowed region in current buffer.

Query user for command to run on temp file specified by variable `TeX-region'.
Temp file created based on `TeX-command-list'.  If master file for document
has header or trailer, these are written to temp file before region itself.
Header is all text before `TeX-header-end'.  Trailer is all text after
`TeX-trailer-start'."

  (interactive)
  (setq TeX-current-process-region-p t)
  (let* ((file (TeX-region-file TeX-default-extension))
         (command (TeX-command-query file))
         (beg (point-min))
         (end (point-max))
         (start (save-restriction (widen) (point-min))))
    (if (nth 4 (assoc command TeX-command-list))
        (TeX-region-create file (buffer-substring beg end)
                           (file-name-nondirectory (buffer-file-name))
                           (count-lines start beg)))
    (TeX-command command 'TeX-region-file)))

;;; Narrow to current environment.

(defun TeX-narrow-to-environment (args)
  "Narrow to current environment \(chapter, [sub]section, document, etc.\).
If point outside an environment, move point forward and narrow.
Narrowing preferences are as follows:

 Default: try narrowing to chapter; if not found, narrow to document.
 one \\[universal-argument]: try narrowing to section; if not found, try above.
 two \\[universal-argument]: try narrowing to subsection; if not found, try above."
  (interactive "p")

  (if (not (string= mode-name "LaTeX"))
      (message "Not a LaTeX buffer")

    (let* ((tail "\\( \t\n\\)*")
           (env
            (cond ((eq args 4)
                   "\\\\section\\|\\\\chapter")
                  ((eq args 16)
                   "\\\\subsection\\|\\\\section\\|\\\\chapter")
                  (t
                   "\\\\chapter\\|\\\\section\\\\subsection")))
           (regexp-beg (concat "^\\(" env "\\)" tail))
           (regexp-end (concat "^\\(" env "\\|\\\\end{document}\\)" tail))
           (beg-doc (concat "^\\\\begin{document}" tail))
           (opoint (point))
           beg end)

      (save-excursion
        (save-restriction
          (save-match-data
            (widen)
            (end-of-line)

            (cond
             ;; Search backwards for chapter and section.
             ((re-search-backward regexp-beg nil t)
              (setq beg (match-beginning 0)))

             ;; Not inside a chapter/section.  Repeat by looking
             ;; forward.
             ((re-search-forward regexp-beg nil t)
              (setq beg (match-beginning 0)))

             ;; Search backwards for document.
             ((re-search-backward beg-doc nil t)
              (setq beg (min (1+ (match-end 0)) (point-max))))

             ;; Repeat by looking forward.
             ((re-search-forward beg-doc nil t)
              (setq beg (min (1+ (match-end 0)) (point-max)))))

            ;; Prepare for search for ending.
            (goto-char (if (match-end 0)
                           (max opoint (match-end 0))
                         opoint))

            ;; Look ahead for ending.
            (if (re-search-forward regexp-end nil t)
                (setq end (match-beginning 0))
              (setq end (point-max)))

            )))

      (if (and beg end)
          (progn
            ;; Echo if narrow-to-region will be moving point.
            (if (< opoint beg)
                (message "Moving forward into environment."))
            (narrow-to-region beg end))))))

(add-hook 'TeX-mode-hook
          (function
           (lambda ()
             (define-key LaTeX-mode-map "\C-c\C-v" 'TeX-command-narrowed-region)
             (define-key LaTeX-mode-map "\C-cnn" 'TeX-narrow-to-environment)
             )))


(provide 'tex-narrow) 

;;; tex-narrow.el ends here.
